"""Pyrion is a full custom GPU, yes you heard me right! a custom GPU. It is under development though!

Shader Script - Cindar Shader Language (CSL)
"""

__version__ = "0.1.0"
__author__ = "AkshuDev"
__name__ = "PyrionGPU"
__shader_language__ = "Cindar"
__shader_info__ = {
    "syntax": "Arris",
    "example": '@inc "stdio.alib"\n\nshader void CSL(in vector3 myvar) {\nreturn null;\n}'
}

