"""Many Prebuilt and Custom Errors (You can create one using this also)"""

import os

import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")