"""C Standard Int"""

from . import *

int8_t = char
int16_t = short
int32_t = Int
int64_t = long
uint8_t = unsigned_char
uint16_t = unsigned_short
uint32_t = unsigned_int
uint64_t = unsigned_long

Int8_t = int8_t
Int16_t = int16_t
Int32_t = int32_t
Int64_t = int64_t
Uint8_t = uint8_t
Uint16_t = uint16_t
Uint32_t = uint32_t
Uint64_t = uint64_t
